/*
 * Copyright 2015 Mike Ryan
 *
 * This file is part of Project Ubertooth.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street,
 * Boston, MA 02110-1301, USA.
 */

#ifndef __EGO_H
#define __EGO_H

#include "ubertooth.h"

typedef enum _ego_mode_t {
	EGO_FOLLOW = 0,
	EGO_CONTINUOUS_RX,
	EGO_JAM,
} ego_mode_t;

void ego_main(ego_mode_t mode);
int rfcat_subcommand(uint16_t cmd, uint8_t *body, int body_len);

#endif /* __EGO_H */
