/*
 * Copyright 2020 Mike Ryan
 *
 * This file is part of Project Ubertooth and is released under the
 * terms of the GPL. Refer to COPYING for more information.
 */

#ifndef __XMAS_H
#define __XMAS_H

void xmas_main(void);

#endif /* __XMAS_H */
